//
// camera.h
//
// The camera class represents a centre of projection and a view frustum.
//

#ifndef CAMERA_H
#define CAMERA_H

#include "ray.h"

class Camera
{
public:
    Camera();
    void rayThrough( double x, double y, ray &r );
    void setEye( const vec3 &eye );
    void setLook( double, double, double, double );
    void setLook( const vec3 &viewDir, const vec3 &upDir );
    void setFOV( double );
    void setAspectRatio( double );

    double getAspectRatio() { return aspectRatio; }
private:
    mat3 m;                     // rotation matrix
    double normalizedHeight;    // dimensions of image place at unit dist from eye
    double aspectRatio;
    
    void update();              // using the above three values calculate look,u,v
    
    vec3 eye;
    vec3 look;                  // direction to look
    vec3 u,v;                   // u and v in the 
};

#endif
