//
// light.h
//
// The Light abstract base class and its concrete subclasses.
//

#ifndef __LIGHT_H__
#define __LIGHT_H__

#include "scene.h"

// You can't create an instance of Light.  Instead, you have to use one
// of the subclasses to get a particular kind of light.

class Light
	: public SceneElement
{
public:
	virtual vec3 getAttenuation( const vec3& P ) const = 0;
	virtual vec3 getColour( const vec3& P ) const = 0;
	virtual vec3 getDirection( const vec3& P ) const = 0;

protected:
	Light( Scene *scene, const vec3& col )
		: SceneElement( scene ), colour( col ) {}

	vec3 		colour;
};

class DirectionalLight
	: public Light
{
public:
	DirectionalLight( Scene *scene, const vec3& orien, const vec3& colour )
		: Light( scene, colour ), orientation( orien ) {}

	virtual vec3 getAttenuation( const vec3& P ) const;
	virtual vec3 getColour( const vec3& P ) const;
	virtual vec3 getDirection( const vec3& P ) const;

protected:
	vec3 		orientation;
};

class PointLight
	: public Light
{
public:
	PointLight( Scene *scene, const vec3& pos, const vec3& colour )
		: Light( scene, colour ), position( pos ) {}
	
	virtual vec3 getAttenuation( const vec3& P ) const;
	virtual vec3 getColour( const vec3& P ) const;
	virtual vec3 getDirection( const vec3& P ) const;

protected:
	vec3 position;
};

#endif // __LIGHT_H__
