//
// material.h
//
// The Material class: a description of the phsyical properties of a surface
// that are used to determine how that surface interacts with light.

#ifndef __MATERIAL_H__
#define __MATERIAL_H__

#include "algebra.h"

class Scene;
class ray;
class isect;

class Material
{
public:
    Material()
        : ke( vec3( 0.0, 0.0, 0.0 ) )
        , ka( vec3( 0.0, 0.0, 0.0 ) )
        , ks( vec3( 0.1, 0.3, 0.9 ) )
        , kd( vec3( 0.2, 0.2, 0.8 ) )
        , kt( vec3( 0.0, 0.0, 0.0 ) )
        , shininess( 0.3 )
		, index( 1.5 ) {}

    Material( const vec3& e, const vec3& a, const vec3& s, 
              const vec3& d, double sh, double in )
        : ke( e ), ka( a ), ks( s ), kd( d ), shininess( sh ), index( in ) {}

	virtual ~Material() {}

	virtual vec3 shadeLocal( Scene *scene, const ray& r, const isect& i ) const;
    
    vec3 ke;                    // emissive
    vec3 ka;                    // ambient
    vec3 ks;                    // specular
    vec3 kd;                    // diffuse
    vec3 kt;                    // transmissive
    
    double shininess;
    double index;               // index of refraction

    
    static const Material zero;
                                // material with zero coeffs for everything
                                // as opposed to the "default" material which is
                                // a pleasant blue.

    Material &
    operator+=( const Material &m )
    {
        ke += m.ke;
        ka += m.ka;
        ks += m.ks;
        kd += m.kd;
        kt += m.kt;
        index += m.index;
        shininess += m.shininess;
        return *this;
    }

    friend Material operator*( double d, Material m );
};

inline Material
operator*( double d, Material m )
{
    m.ke *= d;
    m.ka *= d;
    m.ks *= d;
    m.kd *= d;
    m.kt *= d;
    m.index *= d;
    m.shininess *= d;
    return m;
}

#endif // __MATERIAL_H__
