//
// polymesh.h
//
// The Polymesh class, used to represent polygonal data.
//

#ifndef __POLYMESH_H__
#define __POLYMESH_H__

#include <list>
#include <vector>

#include "ray.h"
#include "material.h"
#include "scene.h"


// Sadly, the polymesh is basically a trimesh for now.

class Polymesh : public MaterialSceneObject
{
    class Triangle
    {
        int ids[3];
    public:
        Triangle( int a, int b, int c )
        {
            ids[0] = a;
            ids[1] = b;
            ids[2] = c;
        }

        int operator[]( int i ) const
        {
            return ids[i];
        }
    };

    typedef vector<vec3> Normals;
    typedef vector<vec3> Vertices;
    typedef vector<Triangle> Faces;
    typedef vector<Material*> Materials;
    Vertices vertices;
    Faces faces;
    Normals normals;
    Materials materials;
public:
    Polymesh( Scene *scene, Material *mat )
        : MaterialSceneObject( scene, mat )
    {
    }

    Polymesh( Scene *scene )
        : MaterialSceneObject( scene )
    {
    }

    ~Polymesh();
    
    virtual bool intersect( const ray& r, isect& i ) const;
    virtual bool intersect( const ray& r ) const;
    bool intersect( const ray& r, const Triangle &tri, isect& i ) const;
    bool intersect( const ray& r, const Triangle &tri ) const;

                                // must add vertices, normals, and materials IN ORDER
    void addVertex( const vec3 & );
    bool addFace( int a, int b, int c );
    void addMaterial( Material *m );
    void addNormal( const vec3 & );

    char *doubleCheck();
    
    void generateNormals();
};

#endif // __POLYMESH_H__
