#include "trace.h"

#include "light.h"
#include "material.h"

vec3
RayTracer::trace( Scene *scene, double x, double y )
// TRACE ray through normalized window coords (x,y)
{
    ray r( vec3(0,0,0), vec3(0,0,0) );
    scene->getCamera()->rayThrough( x,y,r );
    return trace( scene, r );
}

// Trace a top-level ray from the eye, through the projection plane,
// and out into the scene.  All we do is enter the main ray-tracing 
// method, getting things started by plugging in an initial ray weight 
// of (0.0,0.0,0.0) and an initial recursion depth of 0.
vec3 RayTracer::trace( Scene *scene, const ray& r )
{
	return traceRay( scene, r, vec3(1.0,1.0,1.0), 0 );
}

// Do recursive ray tracing!  You'll want to insert a lot of code here
// (or places called from here) to handle reflection, refraction, etc etc.
vec3 RayTracer::traceRay( Scene *scene, const ray& r, 
	const vec3& thresh, int depth )
{
	isect i;

	if( scene->intersect( r, i ) ) {
		// YOUR CODE HERE

		// An intersection occured!  We've got work to do.  For now,
		// this code gets the material for the surface that was intersected,
		// and asks that material to provide a colour for the ray.  

		// This is a great place to insert code for recursive ray tracing.
		// Instead of just returning the result of shadeLocal(), add some
		// more steps: add in the contributions from reflected and refracted
		// rays.

		const Material& m = i.getMaterial();
		return clamp( m.shadeLocal( scene, r, i ) );
	} else {
		// No intersection.  This ray travels to infinity, so we colour
		// it according to the background colour, which in this (simple) case
		// is just black.

		return vec3( 0.0, 0.0, 0.0 );
	}
}
